/** \file Layout.h Declarations for the Layout class.
    \author Liliya Kharevych
    \date March 2006
*/
#ifndef LAYOUT_CM
#define LAYOUT_CM
#include <stack>
#include <vector>
#include "Mesh.h"

/** Class that assigns UV coordinates for mesh vertices

If no cone singularities are used no extra information is required for the layout.

If cone singularities are used, then cuts need to be defined. The cuts need to be defined in such a way that
when mesh is cut, each patch is a disk with single boundary loop and each cone singularity is at 
the boundary of at least one patch. 

The user is responsible for providing valid cuts via -ie option.
*/
class Layout
{
private:
   Mesh * mesh;


   /** Stack of edges that point to the faces that might need to be layout */
   std::stack<Edge *> edgesToFacesTodo;

   /** 
   Given an edge with UV coordinates assigned for the beginning and end vertices of the 
   edge, this function puts the third vertex of the face that it belongs two. 
   */
   void putFaceFrom(Edge * e);


public:
   Layout(Mesh * _mesh);

   /** 
   Computes edge lengths in the final parameterization.

   When radii of circumcircles (r) and intersection angles (theta) are known, final edge length 
   for an edge e can be computed as: \f$l_e = 2 r_1 \sin(\varphi_e)\f$
   
   Where for the internal edges 
   \f$\varphi_e=\textrm{ atan2 }(\sin\theta_e, e^{r_2-r_1} - \cos\theta_e)=\frac{\pi}{2}-\textrm{ atan2 }(1-e^{r_2-r_1}\cos\theta, e^{r_2-r_1}\sin\theta)\f$,
   and for the boundary edges \f$\varphi_e=\pi - \theta_e\f$. 
   
   \f$\varphi_e\f$ is also alphaOposite angle associated with edge e. 

   This function needs to be called before layoutMesh function.
   */
   void computeEdgeLengthAndAngles();

   /** 
   This function assigns UV coordinates based on the information 
   returned from circle pattern optimization. 

   If the mesh is closed, or has
   genus more then 0, cone singularities and cuts need to be given. 
   */
   void layoutMesh();

   /** 
   Moves the patches in the UV plane so that they do not overlap when the mesh is 
   cut into separate components.

   The patches are not packed though in some intelligent way, they are just shifted 
   form each other in x and y direction to approximately cover a square.

   More sophisticated packing could be done as a post processing step.
   */
   void movePatches();

   /** 
   Checks if the cuts given are valid cuts.

   When this function is called, the mesh is already cut.

   First check that each patch is topologically a disk, which means 
   that for each patch the formula: V - E + F = 1 holds.
   Next, check that every cone singularity vertex is on the boundary of the patch.
   */
   bool validCuts();

   ~Layout(void);
};

#endif

